/* main.c */
/*
    This file is part of lwc-gage-avr.
    Copyright (C) 2018 bg nerilex (bg@nerilex.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <string.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <test_vectors.h>
#include "performance_test.h"
#include "avr_uart_i.h"
#include "gage_std.h"

#if DEBUG
#define TEST_MODE 1
#else
#define TEST_MODE 0
#endif

#if 0
int main(void) __attribute__((noreturn));
#endif

#ifndef GAGE_CORE
#error "GAGE_CORE is not defined"
#endif

#ifndef CRYPTO_BYTES
#error "CRYPTO_BYTES is not defined"
#endif


void avr_setup(void);

#if __CHAR_BIT__ != 8
#error "your machines char size is not 8 bits; no luck for you :-("
#endif

#define MIN(a, b) (((a) > (b)) ? (b) : (a))

#if 0
void print_state(const struct state *s) {
    int i;
    printf(" [ ");
    for (i = 0; i < INTERNAL_STATE_SIZE_BYTES - 1; ++i) {
        printf("%02x, ", s->a[i]);
    }
    printf("%02x ]\n", s->a[i]);
}
#endif

static void hex_out_f(FILE *f, const void *buf, size_t len)
{
    if (!f) {
        f = stdout;
    }
    while (len--) {
        printf("%02x", *(unsigned char *)buf);
        buf = 1 + (char *)buf;
    }
}


int crypto_hash(
unsigned char *out,
const unsigned char *in,
unsigned long long inlen
)
{
    struct state st;
    unsigned char mpad[RATE_BYTES];

    Initialize(&st);
//    print_state(&st);
    Padd(mpad, in, inlen);
//    print_state(&st);
    Absorb(&st, in, inlen, mpad);
//    print_state(&st);
    Squeeze(&st, out);
//    print_state(&st);

    return 0;
}

static void test_performance() {
    struct state st;
    uint32_t t;

    Initialize(&st);
    st.a[28] = 0xff;
    st.a[0] = 0xaa;
    uart0_flush();
    startTimer(1);
    MEASURE_b(GAGE_CORE(&st));
    t = stopTimer();
    print_time_P(PSTR("\nGAGE_CORE() Cycles: "), t);

}

#if TEST_MODE == 2
void convert_to_bitslice(struct state *s);
void convert_from_bitslice(struct state *s);
void print_bs_state(const struct state *s, const char *str);

static void bs_test()
{
    unsigned i;
    struct state st;
    uint32_t t;

    Initialize(&st);
#if 1
    for(i = 0; i < INTERNAL_STATE_SIZE_BYTES; ++i) {
        st.a[i] = i;
    }
#else
    i = 0;
    st.a[i++] = 0xaa;
    st.a[i++] = 0xbb;
    st.a[i++] = 0xcc;
    st.a[i++] = 0xdd;
#endif
    for(i = 0; i < sizeof(st.padding); ++i) {
        st.padding[i] = INTERNAL_STATE_SIZE_BYTES + i;
    }
    printf("   non-sliced:");
    print_state(&st);
#if 0
    convert_to_bitslice(&st);
    print_bs_state(&st, "       sliced");
    printf("  pre-convert:");
    print_state(&st);
    convert_from_bitslice(&st);
#else
    GAGE_CORE(&st);
#endif
    printf("re-non-sliced:");
    print_state(&st);
}
#endif

void test_once()
{
    unsigned i;
    uint8_t msg[64] = { 0 };
    uint8_t md[CRYPTO_BYTES];

    struct state st;

    Initialize(&st);
    st.a[0] = 0xA0;
    st.a[28] = 0xff;
    GAGE_CORE(&st);

    for (i = 0; i < sizeof(msg); ++i) {
     msg[i] = (uint8_t)i;
    }
    for (i = 0; i < 1; ++i) {
        printf_P(PSTR("Count = %u\n"), i + 1);
        printf_P(PSTR("Msg = ")); hex_out_f(NULL, msg, i); putchar('\n');
        crypto_hash(md, msg, i);
        printf_P(PSTR("\nMD = ")); hex_out_f(NULL, md, CRYPTO_BYTES); putchar('\n');
        putchar('\n');
    }
}

static void test()
{
    unsigned i;
    uint8_t errors = 0;
    uint8_t msg[64] = { 0 };
    uint8_t md[CRYPTO_BYTES];
    const uint8_t *v = &test_vectors;

    for (i = 0; i < sizeof(msg); ++i) {
        msg[i] = (uint8_t)i;
    }

    for (i = 0; i < sizeof(msg) && i < num_test_vectors; ++i) {
        crypto_hash(md, msg, i);
        if (memcmp_P(md, v, CRYPTO_BYTES) != 0) {
            fputs("!", stdout);
            fflush(NULL);
            ++errors;
        } else {
            fputs("*", stdout);
            fflush(NULL);
        }
        v = &v[CRYPTO_BYTES];
    }

#if 0
    for (i = 0; i < sizeof(msg); ++i) {
        printf_P(PSTR("Count = %u\n"), i + 1);
        printf_P(PSTR("Msg = ")); hex_out_f(NULL, msg, i); putchar('\n');
        crypto_hash(md, msg, i);
        printf_P(PSTR("\nMD = ")); hex_out_f(NULL, md, CRYPTO_BYTES); putchar('\n');
        putchar('\n');
    }
#endif

    putc('\n', stdout);
    if (errors == 0) {
        printf("\n ** OK (%u / %u)\n", num_test_vectors, num_test_vectors);
    } else {
        printf("\n !! FAIL (%u / %u)\n", errors, num_test_vectors);
    }
}


void avr_setup(void);

int main(void)
{
    uint32_t t;
    avr_setup();

    calibrateTimer();
    print_overhead();
    (void)t;
    printf("<INST> hash bytes:  %u\n", CRYPTO_BYTES);
    printf("<INST> state bytes: %u\n", INTERNAL_STATE_SIZE_BYTES);
    printf("<INST> rate bytes:  %u\n", RATE_BYTES);
    uart0_flush();

    printf("Test\n");
#if TEST_MODE == 2
    bs_test();
#elif TEST_MODE == 1
    test_once();
#else
    test();
    putchar('\n');
    test_performance();
#endif

    putchar('\n');

    uart0_flush();
    cli();
    sleep_enable();
    sleep_cpu();
    for (;;) {
    }

    return 0;
}
